from dataclasses import dataclass
from typing import Generator, Union, Generic, TypeVar

from .utils import best_by_key, last, split_lines_into_bounded


P = TypeVar("P")
T = TypeVar("T")
@dataclass
class Success(Generic[P]):
    value: P
    next_index: int


@dataclass
class Failure:
    at_index: int
    expected: set[str]
    children: list["Failure"]

    def __init__(self, at_index: int, expected: set[str], *children: "Failure") -> None:
        self.at_index = at_index
        self.expected = expected
        for child in children:
            for e in child.expected:
                self.expected.add(e)
        self.children = [*children]

    def depth(self):
        depths = [child.depth() + 1 for child in self.children]
        return max(0, 0, *depths)

    def deepest_descendency(self):
        def generator(failure: Failure) -> Generator[Failure, None, None]:
            deepest = best_by_key(failure.children, lambda f: f.depth())
            if deepest is None:
                return
            yield deepest
            yield from generator(deepest)
        return [*generator(self)]

    def parent(self):
        return Failure(self.at_index, self.expected, self)


P = TypeVar("P")
Result = Union[Failure, Success[P]]


@dataclass
class ParseError(BaseException):
    failure: Failure
    stream: str

    def __init__(self, failure: Failure, stream: str) -> None:
        self.failure = failure
        self.stream = stream
        message = self.format_error_secton()
        super().__init__(message)

    def format_error_secton(self):
        actual_failure = last(self.failure.deepest_descendency()) or self.failure
        failure_pos = actual_failure.at_index
        lines = [*split_lines_into_bounded(self.stream)]
        failure_line_index = next(i for (i, ((s, e), l)) in enumerate(lines) if s <= failure_pos and e >= failure_pos)
        result = "\n"
        if failure_line_index > 0:
            index = failure_line_index - 1
            result += format_error_section_line(index, lines[index][1])[1] + "\n"
        (prefix_len, line) = format_error_section_line(failure_line_index, lines[failure_line_index][1])
        result += line + "\n"
        result += format_error_section_indicator(failure_pos, lines[failure_line_index][0][0], prefix_len, actual_failure.expected) + "\n"
        if failure_line_index < (len(lines) - 1):
            index = failure_line_index + 1
            result += format_error_section_line(index, lines[index][1])[1]
        return result


def format_error_section_line(index: int, line: str):
    line_num_part = str(index).rjust(4)
    prefix = f"{line_num_part} |"
    result = f"{prefix}{line}"
    return (len(prefix), result)


def format_error_section_indicator(failure_pos: int, failure_line_start_pos: int, formatted_line_prefix_len: int, expected: set[str]):
    relative_pos = failure_pos - failure_line_start_pos
    padding = formatted_line_prefix_len + relative_pos
    expects = ", ".join(e for e in expected)
    return f"{' ' * padding}^\\_ expected : {expects}"
