from dataclasses import dataclass
from typing import Generic, Optional, TypeVar

from ..result import Result
from ..transformer import Transformer

from .map import MapTransform
from .option import OptionTransform
from .and_ import AndTransform
from .list import ListTransform


T = TypeVar("T")
S = TypeVar("S")
@dataclass
class SepListTransform(Generic[T, S], Transformer[list[T]]):
    items: Transformer[T]
    sep: Transformer[S]

    def __init__(self, items: Transformer[T], sep: Transformer[S]):
        self.items = items
        self.sep = sep
        first_transformer = OptionTransform(items)
        rest_transformer = ListTransform(AndTransform(sep, items))
        total_transformer = AndTransform(first_transformer, rest_transformer)
        self.actual = MapTransform(total_transformer,SepListTransform.mapping)

    def parse(self, stream: str, at_index: int) -> Result[list[T]]:
        return self.actual.parse(stream, at_index)

    @staticmethod
    def mapping(value: tuple[Optional[T], list[tuple[S, T]]]):
        items = list[T]()
        if value[0] is not None:
            items.append(value[0])
        for (_sep, item) in value[1]:
            items.append(item)
        return items


from okipy.lib import test
from .just import JustTransform
from ..result import Result, Success, Failure


@test()
def test_sep_simple(ctx):
    parser = SepListTransform(JustTransform("arbre"), JustTransform(","))
    input = "arbre"
    assert parser.parse(input, 0) == Success(["arbre"], 5)

@test()
def test_sep_alt(ctx):
    parser = SepListTransform(JustTransform("arbre"), JustTransform(","))
    input = "arbre,arbre"
    assert parser.parse(input, 0) == Success(["arbre", "arbre"], 11)

@test()
def test_sep_none(ctx):
    parser = SepListTransform(JustTransform("arbre"), JustTransform(","))
    input = ""
    assert parser.parse(input, 2) == Success([], 2)

@test()
def test_sep_missing_not_failure(ctx):
    parser = SepListTransform(JustTransform("arbre"), JustTransform(","))
    input = "arbre,"
    assert parser.parse(input, 0) == Success(['arbre'], 5)
